#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .combination import MuSyC
from .higher import MuSyC as MuSyC_higher
from .utils import dose_tools
from .utils import base as utils
from .single import Hill


def ds_A(noise=0.05):
    E0, E1, E2, E3 = 1, 0.4, 0.3, 0
    h1, h2 = 2.3, 0.8
    C1, C2 = 0.035, 0.15
    alpha12, alpha21 = 0.3, 4.5
    gamma12, gamma21 = 3.7, 0.95

    model = MuSyC(E0=E0, E1=E1, E2=E2, E3=E3, h1=h1, h2=h2, C1=C1, C2=C2, alpha12=alpha12, alpha21=alpha21, gamma12=gamma12, gamma21=gamma21)
    
    d1, d2 = dose_tools.grid(C1/20, C1*20, C2/20, C2*20, 6, 6, include_zero=True)

    E = model.E(d1, d2)
    noise = noise*(E0-E3)

    np.random.seed(0)
    E = E + noise*(2*np.random.rand(len(d1))-1)
    
    return d1, d2, E

def ds_B(noise=0.05):
    E0, E1, E2, E3 = 1, 0.4, 0.1, 0.2
    h1, h2 = 2.3, 0.8
    C1, C2 = 0.035, 0.15
    alpha12, alpha21 = 1, 2.3
    gamma12, gamma21 = 0.5, 1.02

    model = MuSyC(E0=E0, E1=E1, E2=E2, E3=E3, h1=h1, h2=h2, C1=C1, C2=C2, alpha12=alpha12, alpha21=alpha21, gamma12=gamma12, gamma21=gamma21)
    
    d1, d2 = dose_tools.grid(C1/20, C1*20, C2/20, C2*20, 6, 6, include_zero=True)

    E = model.E(d1, d2)
    noise = noise*(E0-E3)

    np.random.seed(10)
    E = E + noise*(2*np.random.rand(len(d1))-1)
    
    return d1, d2, E

def ds_C(noise=0.05):
    E0, E1, E2, E3 = 1, 0, 0, 0
    h1, h2 = 2.3, 0.5
    C1, C2 = 0.035, 0.15
    alpha12, alpha21 = 0.7, 2.3
    gamma12, gamma21 = 0.9, 1.02

    model = MuSyC(E0=E0, E1=E1, E2=E2, E3=E3, h1=h1, h2=h2, C1=C1, C2=C2, alpha12=alpha12, alpha21=alpha21, gamma12=gamma12, gamma21=gamma21)
    
    d1, d2 = dose_tools.grid(C1/20, C1*20, C2/20, C2*20, 6, 6, include_zero=True)

    E = model.E(d1, d2)
    noise = noise*(E0-E3)

    np.random.seed(20)
    E = E + noise*(2*np.random.rand(len(d1))-1)
    
    return d1, d2, E



def bliss_independent(noise=0.05):
    E0, E1, E2, E3 = 1, 0.4, 0.3, 0.12
    h1, h2 = 2.3, 0.8
    C1, C2 = 0.035, 0.15
    alpha12, alpha21 = 1, 1
    gamma12, gamma21 = 1, 1

    model = MuSyC(E0=E0, E1=E1, E2=E2, E3=E3, h1=h1, h2=h2, C1=C1, C2=C2, alpha12=alpha12, alpha21=alpha21, gamma12=gamma12, gamma21=gamma21)
    
    d1, d2 = dose_tools.grid(C1/20, C1*20, C2/20, C2*20, 6, 6, include_zero=True)

    E = model.E(d1, d2)
    noise = noise*(E0-E3)

    np.random.seed(230)
    E = E + noise*(2*np.random.rand(len(d1))-1)
    
    return d1, d2, E

def linear_isobole(noise=0.05):
    E0, E1, E2, E3 = 1, 0.4, 0., 0
    h1, h2 = 1, 1
    C1, C2 = 0.035, 0.15
    alpha12, alpha21 = 0, 0
    gamma12, gamma21 = 1, 1
    
    model = MuSyC(E0=E0, E1=E1, E2=E2, E3=E3, h1=h1, h2=h2, C1=C1, C2=C2, alpha12=alpha12, alpha21=alpha21, gamma12=gamma12, gamma21=gamma21)
    
    d1, d2 = dose_tools.grid(C1/20, C1*20, C2/20, C2*20, 6, 6, include_zero=True)

    E = model.E(d1, d2)
    noise = noise*(E0-E3)

    np.random.seed(1)
    E = E + noise*(2*np.random.rand(len(d1))-1)
    
    return d1, d2, E

def sham(noise=0.05):
    E0, Emax = 1, 0
    h = 1
    C = 0.035
    drug = Hill(E0=E0, Emax=Emax, h=h, C=C)
    
    d = np.linspace(0,1.5*C,num=10)
    d1, d2, E = utils.sham(d, drug)

    noise = noise*(E0-Emax)
    np.random.seed(22)
    E = E + noise*(2*np.random.rand(len(d1))-1)
    
    return d1, d2, E


# ################### HIGHER

def ds_A_3(noise=0.05):
    E_params = [1,0.5,0.4,0.2,0.3,0.03,0.12,0.1]
    h_params = [2.3,0.8,1.3]
    C_params = [0.1,0.01,0.1]
    alpha_params = [2,3,1,1,0.7,0.5,2,1,1]
    gamma_params = [0.4,2,1,2,0.7,3,2,0.5,2]

    model = MuSyC_higher()
    model.parameters = E_params + h_params + C_params + alpha_params + gamma_params
    
    d = dose_tools.grid_multi((1e-3,1e-3,1e-3),(1,1,1),(6,6,6), include_zero=True)

    E = model.E(d)
    noise = noise*(max(E_params)-min(E_params))

    np.random.seed(64)
    E = E + noise*(2*np.random.rand(d.shape[0])-1)
    
    return d, E

def ds_B_3(noise=0.05):
    E_params = [1,0.45,0.4,0.4,0.3,0.3,0.3,0.]
    h_params = [0.3,0.8,1.3]
    C_params = [0.1,0.01,0.1]
    alpha_params = [2,3,1,1,0.7,0.5,0.2,1,1.5]
    gamma_params = [1.4,2,0.1,2.5,0.7,3,2,0.5,2]

    model = MuSyC_higher()
    model.parameters = E_params + h_params + C_params + alpha_params + gamma_params
    
    d = dose_tools.grid_multi((1e-3,1e-3,1e-3),(1,1,1),(6,6,6), include_zero=True)

    E = model.E(d)
    noise = noise*(max(E_params)-min(E_params))

    np.random.seed(6400)
    E = E + noise*(2*np.random.rand(d.shape[0])-1)
    
    return d, E

def ds_C_3(noise=0.05):
    E_params = [1,0.45,0.4,0.55,0.2,0.1,0.3,0.]
    h_params = [0.3,0.8,1.3]
    C_params = [0.1,0.01,0.1]
    alpha_params = [2,3,0.02,1,2.7,1.5,0.2,0.12,3.7]
    gamma_params = [1.4,2,0.1,2.5,0.7,3,2,0.5,2]

    model = MuSyC_higher()
    model.parameters = E_params + h_params + C_params + alpha_params + gamma_params
    
    d = dose_tools.grid_multi((1e-3,1e-3,1e-3),(1,1,1),(6,6,6), include_zero=True)

    E = model.E(d)
    noise = noise*(max(E_params)-min(E_params))

    np.random.seed(6004)
    E = E + noise*(2*np.random.rand(d.shape[0])-1)
    
    return d, E

def bliss_independent_3(noise=0.05):
    E_params = [1,0.5,0.4,0.2,0.3,0.15,0.12,0.06]
    h_params = [2.3,0.8,1.3]
    C_params = [0.1,0.01,0.1]
    alpha_params = [1,]*9
    gamma_params = [1,]*9

    model = MuSyC_higher()
    model.parameters = E_params + h_params + C_params + alpha_params + gamma_params
    
    d = dose_tools.grid_multi((1e-3,1e-3,1e-3),(1,1,1),(6,6,6), include_zero=True)

    E = model.E(d)
    noise = noise*(max(E_params)-min(E_params))

    np.random.seed(4)
    E = E + noise*(2*np.random.rand(d.shape[0])-1)
    
    return d, E

def linear_isobole_3(noise=0.05):
    E_params = [1,0.5,0.4,0.5,0.2,0,0,0]
    h_params = [1,1,1]
    C_params = [0.1,0.01,0.1]
    alpha_params = [0,]*9
    gamma_params = [1,]*9

    model = MuSyC_higher()
    model.parameters = E_params + h_params + C_params + alpha_params + gamma_params
    
    d = dose_tools.grid_multi((1e-3,1e-3,1e-3),(1,1,1),(6,6,6), include_zero=True)

    with np.errstate(divide='ignore'):
        E = model.E(d)
    noise = noise*(max(E_params)-min(E_params))

    np.random.seed(44)
    E = E + noise*(2*np.random.rand(d.shape[0])-1)
    
    return d, E

def sham_3(noise=0.05):
    E0, Emax = 1, 0
    h = 1
    C = 0.035
    drug = Hill(E0=E0, Emax=Emax, h=h, C=C)
    
    d = np.linspace(0,1.5*C,num=6)
    d, E = utils.sham_higher(d, drug, 3)

    noise = noise*(E0-Emax)
    np.random.seed(288)
    E = E + noise*(2*np.random.rand(d.shape[0])-1)
    
    return d, E


E_params = [1,0.5,0.4,0.5,0.2,0,0,0]
#h_params = [2,1,0.8]
h_params = [1,1,1]
C_params = [0.1,0.01,0.1]
alpha_params = [2,3,1,1,0.7,0.5,2,1,1]
#gamma_params = [0.4,2,1,2,0.7,3,2,0.5,2]
#alpha_params = [0,]*9
gamma_params = [1,]*9

params = E_params + h_params + C_params + alpha_params + gamma_params

truemodel = MuSyC()
truemodel.parameters = params

d = dose_tools.grid_multi((1e-3,1e-3,1e-3),(1,1,1),(6,6,6), include_zero=True)



#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .. import utils
from .parametric_base import ParametricHigher
from ..single import Hill

class MuSyC(ParametricHigher):
    def __init__(self, E_bounds=(-np.inf,np.inf), h_bounds=(0,np.inf), C_bounds=(0,np.inf), alpha_bounds=(0,np.inf), gamma_bounds=(0,np.inf), r=1., parameters=None, variant="full"):
        super().__init__(parameters=parameters)
        
        self.E_bounds = E_bounds
        self.h_bounds = h_bounds
        self.C_bounds = C_bounds
        self.alpha_bounds = alpha_bounds
        self.gamma_bounds = gamma_bounds
        self.r = r

        with np.errstate(divide='ignore'):
            self.logh_bounds = (np.log(h_bounds[0]), np.log(h_bounds[1]))
            self.logC_bounds = (np.log(C_bounds[0]), np.log(C_bounds[1]))
            self.logalpha_bounds = (np.log(alpha_bounds[0]), np.log(alpha_bounds[1]))
            self.loggamma_bounds = (np.log(gamma_bounds[0]), np.log(gamma_bounds[1]))

        # Bounds will depened on the number of dimensions, so will be filled out in _get_initial_guess()

        self.variant = variant

        if variant == "no_gamma":
            self.fit_function = self._model_no_gamma
        else:
            self.fit_function = self._model

        # Given 3 drugs, there are 9 synergy edges, and so 9 synergistic potencies and cooperativities. Thus, alphas=[a,b,c,d,e,f,g,h,i]. _edge_index[2][6] will give the index of the alpha corresponding to going from state [010] to [110] (e.g., adding drug 3).
        self._edge_index = None

    def fit(self, d, E, bootstrap_iterations=0, **kwargs):
        if len(d.shape) != 2:
            return None
        
        self._build_edge_indices(d.shape[1])
        super().fit(d, E, bootstrap_iterations=bootstrap_iterations, **kwargs)

    def E(self, d):
        if len(d.shape) != 2:
            # d is not properly formatted
            return None

        n = d.shape[1]
        if (n < 2):
            # MuSyC requires at least two drugs
            return None

        if not self._is_parameterized():
            return None

        self._build_edge_indices(n)
        params = self._transform_params_to_fit(self.parameters)

        if self.variant == "no_gamma":
            return self._model_no_gamma(d, *params)
        return self._model(d, *params)

    def _transform_params_to_fit(self, params):
        params = np.array(params, copy=True)
        n = self._get_n_drugs_from_params(params)
        if n<2:
            return None
        
        h_param_offset = 2**n
        params[h_param_offset:] = np.log(params[h_param_offset:])
        return params

    def _transform_params_from_fit(self, popt):
        params = np.array(popt, copy=True)
        n = self._get_n_drugs_from_params(params)
        if n<2:
            return None
        
        h_param_offset = 2**n
        params[h_param_offset:] = np.exp(params[h_param_offset:])
        return params

    def _get_n_drugs_from_params(self, params):
        n = 2
        if self.variant == "no_gamma":
            threshold = lambda n : 2**n + n + 2**(n-1)*n
        else:
            threshold = lambda n : 2**n + n*2**n
        while len(params) > threshold(n):
            n += 1
        if len(params) == threshold(n):
            return n
        return 0

    def _get_initial_guess(self, d, E, p0=None):
        n = d.shape[1]

        n_E = 2**n
        n_h = n
        n_C = n
        n_alpha = 2**(n-1)*n-n
        n_gamma = 2**(n-1)*n-n

        E_bounds = [self.E_bounds,]*n_E
        logh_bounds = [self.logh_bounds,]*n_h
        logC_bounds = [self.logC_bounds,]*n_C
        logalpha_bounds = [self.logalpha_bounds,]*n_alpha
        
        if self.variant == "no_gamma":
            bounds = E_bounds + logh_bounds + logC_bounds + logalpha_bounds
        else:
            loggamma_bounds = [self.loggamma_bounds,]*n_gamma
            bounds = E_bounds + logh_bounds + logC_bounds + logalpha_bounds + loggamma_bounds

        self.bounds = tuple(zip(*bounds))
        
        if p0 is None:
            E_params = [0,]*(2**n)
            h_params = [1,]*n
            C_params = [np.median(d[:,i]) for i in range(n)]
            alpha_params = [1,]*(2**(n-1)*n-n)
            gamma_params = [1,]*(2**(n-1)*n-n)
        
            # Make guesses of E for each drug state
            for idx in range(2**n):
                # state = [0,1,1] means drug3=0, drug2=1, drug1=1
                state = MuSyC._idx_to_state(idx, n)
                mask = d[:,0]>-1 # d is always > 0, so initializes to array of True
                for drugnum in range(1,n+1): # 1, 2, 3, ...
                    drugstate = state[n-drugnum] # 1->2, 2->1, 3->0
                    if drugstate==0:
                        mask = mask & (d[:,drugnum-1]==np.min(d[:,drugnum-1]))
                    else:
                        mask = mask & (d[:,drugnum-1]==np.max(d[:,drugnum-1]))
                mask = np.where(mask)
                E_params[idx] = np.median(E[mask])

            # Make guesses for E, h, C of undrugged and single-drugged states
            single_drug_model = Hill(E0_bounds=self.E_bounds, Emax_bounds=self.E_bounds, h_bounds=self.h_bounds, C_bounds=self.C_bounds)
            E0_guess = 0

            # Make Hill model of each single drug
            for i in range(n):
                # Mask all other drugs at their minimum values
                mask = d[:,0]>-1 # d is always > 0, so initializes to array of True
                for otherdrug in range(n):
                    if otherdrug==i:
                        continue
                    mask = mask & (d[:,otherdrug] == np.min(d[:,otherdrug]))
                mask = np.where(mask)
                single_drug_model.fit(d[mask,i].flatten(), E[mask], p0=(E_params[0], E_params[i], h_params[i], C_params[i]))

                # Override initial guesses with single fits
                if single_drug_model.converged:
                    E0_guess += single_drug_model.E0/n
                    E_params[i] = single_drug_model.Emax
                    h_params[i] = single_drug_model.h
                    C_params[i] = single_drug_model.C
                else:
                    E0_guess += E_params[0]/n
            E_params[0] = E0_guess
            
            if self.variant == "no_gamma":
                p0 = E_params + h_params + C_params + alpha_params
            else:
                p0 = E_params + h_params + C_params + alpha_params + gamma_params
        
        p0 = list(self._transform_params_to_fit(p0))
        utils.sanitize_initial_guess(p0, self.bounds)
        return p0
        
    @staticmethod
    def _hamming(a,b):
        s = 0
        for A,B in zip(a,b):
            if (A!=B): s+=1
        return s

    @staticmethod
    def _get_n_states(n):
        return 2**n

    @staticmethod
    def _get_n_edges(n):
        return 2**(n-1)*n

    @staticmethod
    def _idx_to_state(idx, n):
        """Returns state representing index=idx

        MuSyC models states for cells affected by each drug in all possible combinations. For instance, given three drugs, MuSyC will have states 000, 001, 010, 011, 100, 101, 110, 111. 110 means affected by drugs 3 and 2, but unaffected by drug 1.

        Parameters
        ----------
        idx : int
            Index for the state, ranging from 0 to 2**n-1.
        
        n : int
            Number of drugs in combination

        Returns
        -------
        state : list of int
            A list indicating which drugs are "active". [1,1,0] means drug 1 is not active (0), and drugs 2 and 3 are active.
        """
        return [int(i) for i in bin(idx)[2:].zfill(n)]

    @staticmethod
    def _state_to_idx(state):
        """Converts a drug state to its index.
        See _idx_to_state() for more info.

        Parameters
        ----------
        state : list of int
            
        Returns
        -------
        idx : int
        """
        return int(''.join([str(i) for i in state]), base=2)

    @staticmethod
    def _get_neighbors(idx, n):
        """Returns neighbors of the drug.

        Returns
        -------
        add_drugs : list
            List of tuples. For each tuple, the first element indicates \
            which drug is being added. The second element indicates the \
            idx of the state reached by adding that drug.

        remove_drugs : list
            As above, but for states reached by removing drugs.
        """
        state = MuSyC._idx_to_state(idx, n)
        add_drugs = []
        remove_drugs = []
        for i,val in enumerate(state):
            neighbor = [j for j in state]
            neighbor[i] = 1-val
            drug_idx = n-i-1
            if val==0:
                add_drugs.append((drug_idx, MuSyC._state_to_idx(neighbor)))
            else:
                remove_drugs.append((drug_idx, MuSyC._state_to_idx(neighbor)))
        return add_drugs, remove_drugs
    
    def _build_edge_indices(self, n):
        self._edge_index = dict()
        count = 0
        for i in range(1,2**n): 
            add_d, rem_d = MuSyC._get_neighbors(i,n)
            if len(add_d) > 0:
                self._edge_index[i] = dict()
            for j in add_d:
                self._edge_index[i][j[1]] = count
                count += 1

    def _model_no_gamma(self, doses, *args):
        n = doses.shape[1]
        loggammas = [0,]*(2**(n-1)*n-n)
        return self._model(doses, *args, *loggammas)

    def _model(self, doses, *args):
        n = doses.shape[1]
        matrix = np.zeros((doses.shape[0],2**n,2**n))
        
        h_param_offset = 2**n
        C_param_offset = h_param_offset + n
        alpha_param_offset = C_param_offset + n
        gamma_param_offset = alpha_param_offset + 2**(n-1)*n-n

        E_params = args[:2**n]
        logh_params = np.asarray(args[h_param_offset:C_param_offset])
        logC_params = np.asarray(args[C_param_offset:alpha_param_offset])
        logalpha_params = np.asarray(args[alpha_param_offset:gamma_param_offset])
        loggamma_params = np.asarray(args[gamma_param_offset:])

        h_params = np.exp(logh_params)
        C_params = np.exp(logC_params)
        alpha_params = np.exp(logalpha_params)
        gamma_params = np.exp(loggamma_params)

        # First do row 0, which corresponds to U
        # All edges are non-synergistic

        add_drugs, remove_drugs = MuSyC._get_neighbors(0, n)
        for drugnum, jidx in add_drugs:
            d = doses[:,drugnum]
            h = h_params[drugnum]
            C = C_params[drugnum]
            r1r = self.r*np.power(C,h)
            
            matrix[:,0,0] -= self.r*np.power(d,h)
            matrix[:,0,jidx] = r1r

        # Loop over all other states/rows (except the last one)
        for idx in range(1,2**n-1):
            row = [0,]*(2**n)
            add_drugs, remove_drugs = MuSyC._get_neighbors(idx, n)
            for drugnum, jidx in add_drugs:
                gamma = 1
                alpha = 1
                # If the neighbor is not 0, this is a synergy edge
                if jidx > 0:
                    edge_idx = self._edge_index[idx][jidx]
                    gamma = gamma_params[edge_idx]
                    alpha = alpha_params[edge_idx]
                d = doses[:,drugnum]
                h = h_params[drugnum]
                C = C_params[drugnum]
                r1r = self.r*np.power(C,h)
                
                # This state gains from reverse transitions out of jidx
                matrix[:,idx,jidx] += r1r**gamma

                # This state loses from transitions toward jidx
                matrix[:,idx,idx] -= np.power(self.r*np.power(alpha*d,h),gamma)

            for drugnum, jidx in remove_drugs:
                gamma = 1
                alpha = 1
                # If the neighbor is not 0, this is a synergy edge
                if jidx > 0:
                    edge_idx = self._edge_index[jidx][idx]
                    gamma = gamma_params[edge_idx]
                    alpha = alpha_params[edge_idx]
                d = doses[:,drugnum]
                h = h_params[drugnum]
                C = C_params[drugnum]
                r1r = self.r*np.power(C,h)
                
                # This state loses from reverse transitions toward jidx
                matrix[:, idx, idx] -= r1r**gamma

                # This state gaines from transitions from jidx
                matrix[:, idx, jidx] += np.power(self.r*np.power(alpha*d,h),gamma)
            
        # The final constraint is that U+A1+A2+... = 1
        matrix[:,-1,:]=1
        matrix_inv = np.linalg.inv(matrix)

        # All other rows should multiply to zero. Only the last row goes to 1.
        b = np.zeros(2**n)
        b[-1]=1
        return np.dot(np.dot(matrix_inv,b), np.asarray(E_params))

    @staticmethod
    def _state_to_drugstr(s, sb=None):
        """Converts state (e.g., [1,1,0]) to drug-string (e.g., "2,3")
        """
        # If state is undrugged, return 0
        if s.count(1)==0:
            return 0
        n = len(s)
        snums = []
        for i in range(n):
            if s[n-i-1]==1:
                snums.append(str(i+1))
        if sb is None:
            return ','.join(snums)
        sbnums = []
        for i in range(n):
            if sb[n-i-1]==1:
                sbnums.append(str(i+1))
        sbnums = [i for i in sbnums if not i in snums]
        return ','.join(snums), ','.join(sbnums)

    @staticmethod
    def _get_beta(state, parameters):
        """Calculates synergistic efficacy, a synergy parameter derived from E parameters.
        """

        # beta is only defined for states associated with 2 or more drugs
        if state.count(1)<2:
            return 0

        n = len(state)

        idx = MuSyC._state_to_idx(state)
        E = parameters[idx]

        # parents are states with one fewer drug active (e.g., parents of 011 are 001 and 010). beta is calculated by comparing E, the strongest of E_parents, and E0.
        E_parents = []
        for i in range(n):
            if state[i]==1:
                parent = [j for j in state]
                parent[i] = 0
                pidx = MuSyC._state_to_idx(parent)
                E_parents.append(parameters[pidx])

        E0 = parameters[0]
        # TODO Add support for positive E orientation
        E_best_parent = np.amin(np.asarray(E_parents), axis=0)

        beta = (E_best_parent-E)/(E0-E_best_parent)
        return beta


    def get_parameters(self, confidence_interval=95):
        if not self._is_parameterized():
            return None
        
        n = self._get_n_drugs_from_params(self.parameters)
        h_param_offset = 2**n
        C_param_offset = h_param_offset + n
        alpha_param_offset = C_param_offset + n
        gamma_param_offset = alpha_param_offset + 2**(n-1)*n-n
        
        if self.converged and self.bootstrap_parameters is not None:
            parameter_ranges = self.get_parameter_range(confidence_interval=confidence_interval)
        else:
            parameter_ranges = None

        params = dict()

        # Add E parameters
        for idx in range(2**n):
            state = MuSyC._idx_to_state(idx, n)
            #state = "".join([str(s) for s in state])
            statestr = MuSyC._state_to_drugstr(state)
            l = []
            l.append(self.parameters[idx])
            if parameter_ranges is not None:
                l.append(parameter_ranges[:,idx])
            params["E_%s"%statestr] = l

            # Add beta parameters
            if state.count(1) > 1:
                beta = MuSyC._get_beta(state, self.parameters)
                l = []
                l.append(beta)

                if parameter_ranges is not None:
                    beta_bootstrap = MuSyC._get_beta(state, self.bootstrap_parameters.T)

                    beta_bootstrap = np.percentile(beta_bootstrap, [(100-confidence_interval)/2, 50+confidence_interval/2])
                    l.append(beta_bootstrap)
                params["beta_%s"%statestr] = l


        # h and C
        for i in range(n):
            lh = []
            lC = []
            lh.append(self.parameters[h_param_offset+i])
            lC.append(self.parameters[C_param_offset+i])
            if parameter_ranges is not None:
                lh.append(parameter_ranges[:,h_param_offset+i])
                lC.append(parameter_ranges[:,C_param_offset+i])
            params["h%d"%i] = lh
            params["C%d"%i] = lC
        
        # alpha and gamma

        for idx_a in self._edge_index.keys():
            state_a = MuSyC._idx_to_state(idx_a,n)
            
            for idx_b in self._edge_index[idx_a].keys():
                state_b = MuSyC._idx_to_state(idx_b,n)
                state_a_str, state_b_str = MuSyC._state_to_drugstr(state_a, state_b)

                lalpha = []
                lgamma = []

                i = self._edge_index[idx_a][idx_b]
                lalpha.append(self.parameters[alpha_param_offset + i])
                if self.variant != "no_gamma":
                    lgamma.append(self.parameters[gamma_param_offset + i])

                if parameter_ranges is not None:
                    lalpha.append(parameter_ranges[:,alpha_param_offset+i])
                    if self.variant != "no_gamma":
                        lgamma.append(parameter_ranges[:,gamma_param_offset+i])

                params["alpha_%s_%s"%(state_a_str, state_b_str)] = lalpha
                if self.variant != "no_gamma":
                    params["gamma_%s_%s"%(state_a_str, state_b_str)] = lgamma
        
        return params

    def summary(self, confidence_interval=95, tol=0.01):
        pars = self.get_parameters(confidence_interval=confidence_interval)
        if pars is None:
            return None
        
        ret = []
        keys = pars.keys()
        # beta
        for key in keys:
            if "beta" in key:
                l = pars[key]
                if len(l)==1:
                    if l[0] < -tol:
                        ret.append("%s\t%0.2f\t(<0) antagonistic"%(key, l[0]))
                    elif l[0] > tol:
                        ret.append("%s\t%0.2f\t(>0) synergistic"%(key, l[0]))
                else:
                    v = l[0]
                    lb,ub = l[1]
                    if v < -tol and lb < -tol and ub < -tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(<0) antagonistic"%(key, v,lb,ub))
                    elif v > tol and lb > tol and ub > tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(>0) synergistic"%(key, v,lb,ub))
        # alpha
        for key in keys:
            if "alpha" in key:
                l = pars[key]
                if len(l)==1:
                    if np.log10(l[0]) < -tol:
                        ret.append("%s\t%0.2f\t(<1) antagonistic"%(key, l[0]))
                    elif np.log10(l[0]) > tol:
                        ret.append("%s\t%0.2f\t(>1) synergistic"%(key, l[0]))
                else:
                    v = l[0]
                    lb,ub = l[1]
                    if np.log10(v) < -tol and np.log10(lb) < -tol and np.log10(ub) < -tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(<1) antagonistic"%(key, v,lb,ub))
                    elif np.log10(v) > tol and np.log10(lb) > tol and np.log10(ub) > tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(>1) synergistic"%(key, v,lb,ub))

        # gamma
        for key in keys:
            if "gamma" in key:
                l = pars[key]
                if len(l)==1:
                    if np.log10(l[0]) < -tol:
                        ret.append("%s\t%0.2f\t(<1) antagonistic"%(key, l[0]))
                    elif np.log10(l[0]) > tol:
                        ret.append("%s\t%0.2f\t(>1) synergistic"%(key, l[0]))
                else:
                    v = l[0]
                    lb,ub = l[1]
                    if np.log10(v) < -tol and np.log10(lb) < -tol and np.log10(ub) < -tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(<1) antagonistic"%(key, v,lb,ub))
                    elif np.log10(v) > tol and np.log10(lb) > tol and np.log10(ub) > tol:
                        ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(>1) synergistic"%(key, v,lb,ub))
        
        if len(ret)>0:
            return "\n".join(ret)
        else:
            return "No synergy or antagonism detected with %d percent confidence interval"%(int(confidence_interval))





__version__="0.3.2"
__author__="David J. Wooten"
__citation__="Wooten, DJ and Albert, RA. synergy: A Python library for calculating drug combination synergy. In preparation."

from . import combination
from . import single
from . import utils


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

class MarginalLinear:
    """Some drugs' dose response may not follow some known parametric equation. In these cases, MarginalLinear() fits a piecewise linear model mapping log(d) -> E

    Parameters
    ----------
    aggregation_function : function, default=np.mean
        If d contains repeated values (e.g., experimental replicates using the same dose many times), E at these repeated doses will be averaged using aggregation_function(E[d==X])
    """
    def __init__(self, aggregation_function=np.mean):
        self._d = None
        self._E = None
        self._logd = None
        self._aggregation_function = aggregation_function

    def fit(self, d, E):
        """Calls __init__(d, E, aggregation_function)

        Parameters
        ----------
        d : array_like
            Array of doses measured
        
        E : array_like
            Array of effects measured at doses d
        """
        
        if (len(d) > len(np.unique(d))):
            self._d = []
            self._E = []

            # Given repeated dose measurements, average E for each
            for val in np.unique(d):
                self._d.append(val)
                self._E.append(self._aggregation_function(E[d==val]))

            self._d = np.asarray(self._d)
            self._E = np.asarray(self._E)

        else:
            self._d = np.array(d,copy=True)
            self._E = np.array(E,copy=True)

        # Replace 0 doses with the minimum float value
        self._d[self._d==0] = np.nextafter(0,1)

        # Sort doses and E
        sorted_indices = np.argsort(self._d)
        self._d = self._d[sorted_indices]
        self._E = self._E[sorted_indices]
        
        # Get log-transformed dose (used for interpolation)
        self._logd = np.log(self._d)

    def E(self, d):
        """Evaluate this model at dose d

        Parameters
        ----------
        d : array_like
            Doses to calculate effect at
        
        Returns
        ----------
        effect : array_like
            Evaluate's the model at dose in d
        """
        d = np.array(d, copy=True)
        d[d==0] = np.nextafter(0,1)
        logd = np.log(d)

        # These doses are below the minimum dose, or above the maximum, and therefore cannot be used for interpolation
        invalid_mask = np.where((logd < min(self._logd)) | (logd > max(self._logd)))

        E = np.interp(logd, self._logd, self._E)

        E[invalid_mask] = np.nan

        return E

    def E_inv(self, E):
        """Find the dose that will achieve effects in E. Only works for dose responses with strictly increasing or strictly decreasing E.

        Parameters
        ----------
        E : array_like
            Effects to get the doses for
        
        Returns
        ----------
        doses : array_like
            Doses which achieve effects E using this model.
        """
        if not self._is_invertible():
            ret = 0*E
            ret[:]=np.nan
            return ret
        
        # These effects are below the minimum or above the maximum, and therefore cannot be used for interpolation
        invalid_mask = np.where((E < min(self._E)) | (E > max(self._E)))

        d = np.exp(np.interp(E, self._E, self._logd))

        d[invalid_mask] = np.nan

        return d

    def _is_invertible(self):
        """The dose model is only invertible if E is strictly increasing or strictly decreasing
        """
        
        if (len(self._E) > 1):
            consecutive_diffs = np.ediff1d(self._E)
            if np.all(consecutive_diffs>0) or np.all(consecutive_diffs<0):
                return True
        return False


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ..single import Hill, Hill_2P
from .nonparametric_base import DoseDependentModel

class ZIP(DoseDependentModel):
    """The Zero Interaction Potency (ZIP) model (doi: 10.1016/j.csbj.2015.09.001). This model is based on the multiplicative survival principal (i.e., Bliss). All across the dose-response surface, it fits Hill-equations either holding d1==constant or d2==constant.
    
    ZIP quantifies changes in the EC50 (C) and Hill-slope (h) of each drug, as the other is increased. For instance, at a given d1==D1, d2==D2, ZIP fits two Hill equations: one for drug1 (fixing d2==D2), and one for drug2 (fixing d1==D1). These Hill equations are averaged to get a fit value of E (y_c in their paper) at these doses. This is then subtracted from the expected result (y_zip in their paper) obtained by assuming these Hill equations have equivalent h and C to their single-drug counterparts. This difference (delta in their paper) becomes the metric of synergy.

    ZIP models store these delta values as model._synergy, but also store the Hill equation fits for drug1 and drug2 across the whole surface, allowing investigation of how h and C change across the surface

    synergy : array_like
        (-inf,0)=antagonism, (0,inf)=synergism. The "delta" synergy score from ZIP

    _h_21 : array_like
        The hill slope of drug 1 obtained by holding D2==constant

    _h_12 : array_like
        The hill slope of drug 2 obtained by holding D1==constant

    _C_21 : array_like
        The EC50 of drug 1 obtained by holding D2==constant

    _C_12 : array_like
        The EC50 of drug 2 obtained by holding D1==constant
    """
    def __init__(self, E0_bounds=(-np.inf,np.inf), Emax_bounds=(-np.inf,np.inf), h_bounds=(0,np.inf), C_bounds=(0,np.inf)):

        super().__init__(h1_bounds=h_bounds, h2_bounds=h_bounds, C1_bounds=C_bounds, C2_bounds=C_bounds, E0_bounds=E0_bounds, E1_bounds=Emax_bounds, E2_bounds=Emax_bounds)

        self.E0_bounds = E0_bounds
        self.Emax_bounds = Emax_bounds
        self.C_bounds = C_bounds
        self.h_bounds = h_bounds

        self._h_21 = []
        self._h_12 = []
        self._C_21 = []
        self._C_12 = []

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, use_jacobian=True, **kwargs):
    
        d1 = np.asarray(d1)
        d2 = np.asarray(d2)
        E = np.asarray(E)
        super().fit(d1,d2,E)
        
        if drug1_model is None:
            mask = np.where(d2==min(d2))
            drug1_model = Hill.create_fit(d1[mask], E[mask], E0_bounds=self.E0_bounds, Emax_bounds=self.Emax_bounds, h_bounds=self.h_bounds, C_bounds=self.C_bounds, use_jacobian=use_jacobian)
            
        if drug2_model is None:
            mask = np.where(d1==min(d1))
            drug2_model = Hill.create_fit(d2[mask], E[mask], E0_bounds=self.E0_bounds, Emax_bounds=self.Emax_bounds,h_bounds=self.h_bounds, C_bounds=self.C_bounds, use_jacobian=use_jacobian)

        self.drug1_model = drug1_model
        self.drug2_model = drug2_model
            
        E0_1, Emax_1, h1, C1 = drug1_model.get_parameters()
        E0_2, Emax_2, h2, C2 = drug2_model.get_parameters()
        E0 = (E0_1+E0_2)/2.
        Emax = (Emax_1+Emax_2)/2.
        if (E0 < Emax):
            drug1_model.E0 = Emax_1
            drug2_model.E0 = Emax_2
            drug1_model.Emax = E0_1
            drug2_model.Emax = E0_2

            tmp = E0
            E0 = Emax
            Emax = tmp
        
        logh1 = np.log(h1)
        logh2 = np.log(h2)
        logC1 = np.log(C1)
        logC2 = np.log(C2)

        self._h_21 = [] # Hill slope of drug 1, after treated by drug 2
        self._h_12 = [] # Hill slope of drug 2, after treated by drug 1
        self._C_21 = [] # EC50 of drug 1, after treated by drug 2
        self._C_12 = [] # EC50 of drug 2, after treated by drug 1
        
        zip_model = Hill_2P(Emax=Emax, h_bounds=self.h_bounds, C_bounds=self.C_bounds)

        for D1, D2 in zip(d1, d2):
            # Fix d2==D2, and fit hill for D1
            mask = np.where(d2==D2)
            y2 = drug2_model.E(D2)
            zip_model.E0 = y2
            zip_model.fit(d1[mask],E[mask], use_jacobian=use_jacobian, p0=[h1,C1])
            self._h_21.append(zip_model.h)
            self._C_21.append(zip_model.C)

            # Fix d1==D1, and fit hill for D2
            mask = np.where(d1==D1)
            y1 = drug1_model.E(D1)
            zip_model.E0 = y1
            zip_model.fit(d2[mask],E[mask], use_jacobian=use_jacobian, p0=[h2,C2])
            self._h_12.append(zip_model.h)
            self._C_12.append(zip_model.C)
        
        self._h_21 = np.asarray(self._h_21)
        self._h_12 = np.asarray(self._h_12)
        self._C_21 = np.asarray(self._C_21)
        self._C_12 = np.asarray(self._C_12)

        self.synergy = self._delta_score(d1, d2, E0, Emax, h1, h2, C1, C2, self._h_21, self._h_12, self._C_21, self._C_12)

        return self.synergy

    def _delta_score(self, d1, d2, E0, Emax, h1, h2, C1, C2, h_21, h_12, C_21, C_12):
        dCh1 = np.power(d1/C1,h1)
        dCh2 = np.power(d2/C2,h2)
        dCh1_prime = np.power(d1/C_21,h_21)
        dCh2_prime = np.power(d2/C_12,h_12)
        
        AA = ((E0 + Emax*dCh2)/(1.+dCh2) + Emax*dCh1_prime) / (1+dCh1_prime)
        BB = ((E0 + Emax*dCh1)/(1.+dCh1) + Emax*dCh2_prime) / (1+dCh2_prime)
        #CC = (E0 + Emax*dCh1)/(1+dCh1) + (E0 + Emax*dCh2)/(1+dCh2) - (E0 + Emax*dCh1)/(1+dCh1)*(E0 + Emax*dCh2)/(1+dCh2) # This form expects E0==0, Emax=1
        CC = (E0 + Emax*dCh1)/(1+dCh1)*(E0 + Emax*dCh2)/(1+dCh2)/E0
        # ^^^ The /E0 at the end is needed in case E0 != 1, as AA and BB are on the scale of E0, but CC would have been on the scale of E0^2. The final result is then on the order of E0. So if E ranges from 100 to 0, it will be 100 times greater than if E had ranged from 1 to 0.

        return CC - (AA+BB)/2.



#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import norm

from .. import utils
from ..utils import plots

class ParametricHigher(ABC):
    """The abstract base class for higher dimensional (3+ drug) parametric synergy models.
    """
    def __init__(self, parameters=None):
        self.bounds = None
        self.fit_function = None
        
        self.converged = False
        self.parameters = parameters

        self.sum_of_squares_residuals = None
        self.r_squared = None
        self.aic = None
        self.bic = None
        self.bootstrap_parameters = None

    def _internal_fit(self, d, E, **kwargs):
        """Internal method to fit the model to data (d,E)
        """
        try:
            popt, pcov = curve_fit(self.fit_function, d, E, bounds=self.bounds, **kwargs)
            return self._transform_params_from_fit(popt)
        except Exception as e:
            print("Exception:", e)
            return None

    def fit(self, d, E, bootstrap_iterations=0, **kwargs):
        """Fits the model to data

        Parameters
        ----------
        d : numpy.ndarray (M x N)
            Doses of N drugs sampled at M points
        
        E : array_like
            Dose-response at doses d

        bootstrap_iterations : int , default=0
            The number of boostrap resample iterations performed to estimate parameter confidence intervals.

        kwargs
            kwargs to pass to scipy.optimize.curve_fit()
        """
        d = np.asarray(d)
        E = np.asarray(E)

        if 'p0' in kwargs:
            p0 = list(kwargs.get('p0'))
        else:
            p0 = None
        p0 = self._get_initial_guess(d, E, p0=p0)
        kwargs['p0'] = p0

        with np.errstate(divide='ignore', invalid='ignore'):
            popt = self._internal_fit(d, E, **kwargs)
        
        if popt is None:
            self.converged = False
            self.parameters = self._transform_params_from_fit(p0)

        else:
            self.converged = True
            self.parameters = popt
            n_parameters = len(popt)
            n_samples = d.shape[0]
            if (n_samples - n_parameters - 1 > 0):
                self._score(d, E)
                kwargs['p0'] = self._transform_params_to_fit(popt)
                self._bootstrap_resample(d, E, bootstrap_iterations, **kwargs)
            

    def _bootstrap_resample(self, d, E, bootstrap_iterations, **kwargs):
        """Internal function to identify confidence intervals for parameters
        """

        if not self._is_parameterized(): return
        if not self.converged: return

        n_data_points = len(E)
        n_parameters = len(self.parameters)
        
        sigma_residuals = np.sqrt(self.sum_of_squares_residuals / (n_data_points - n_parameters))

        E_model = self.E(d)
        bootstrap_parameters = []

        for iteration in range(bootstrap_iterations):
            print("Bootstrap iteration = %d"%iteration)
            residuals_step = norm.rvs(loc=0, scale=sigma_residuals, size=n_data_points)

            # Add random noise to model prediction
            E_iteration = E_model + residuals_step

            # Fit noisy data
            with np.errstate(divide='ignore', invalid='ignore', over='ignore'):
                popt1 = self._internal_fit(d, E_iteration, **kwargs)
            if popt1 is not None:
                bootstrap_parameters.append(popt1)

        if len(bootstrap_parameters) > 0:
            self.bootstrap_parameters = np.vstack(bootstrap_parameters)
        else:
            self.bootstrap_parameters = None

    def _score(self, d, E):
        """Calculate goodness of fit and model quality scores, including sum-of-squares residuals, R^2, Akaike Information Criterion (AIC), and Bayesian Information Criterion (BIC).

        If model is not yet paramterized, does nothing

        Called automatically during model.fit(d, E)

        Parameters
        ----------
        d : numpy.ndarray (M x N)
            Doses of N drugs sampled at M points
        
        E : array_like
            Dose-response at doses d
        """
        if (self._is_parameterized()):

            n_parameters = len(self.parameters)

            self.sum_of_squares_residuals = utils.residual_ss_1d(d, E, self.E)
            self.r_squared = utils.r_squared(E, self.sum_of_squares_residuals)
            self.aic = utils.AIC(self.sum_of_squares_residuals, n_parameters, len(E))
            self.bic = utils.BIC(self.sum_of_squares_residuals, n_parameters, len(E))

    def get_parameter_range(self, confidence_interval=95):
        """Returns the lower bound and upper bound estimate for each parameter.

        Parameters:
        -----------
        confidence_interval : int, float, default=95
            % confidence interval to return. Must be between 0 and 100.
        """
        if not self._is_parameterized():
            return None
        if not self.converged:
            return None
        if confidence_interval < 0 or confidence_interval > 100:
            return None
        if self.bootstrap_parameters is None:
            return None

        lb = (100-confidence_interval)/2.
        ub = 100-lb
        return np.percentile(self.bootstrap_parameters, [lb, ub], axis=0)

    def _is_parameterized(self):
        """Returns False if any parameters are None or nan.

        Returns
        ----------
        is_parameterzed : bool
            True if all of the parameters are set. False if any are None or nan.
        """
        return not ((self.parameters is None) or (None in self.parameters) or (True in np.isnan(np.asarray(self.parameters))))

    @abstractmethod
    def _get_initial_guess(self, d, E, p0=None):
        """Internal method to format and/or come up with initial guess for parameters
        """
        pass

    @abstractmethod
    def E(self, d):
        """Evaluates the model at dose d

        Parameters
        ----------
        d : numpy.ndarray
            Doses, in an M x N ndarray, where M is the number of samples, and N is the number of drugs.

        Returns
        ----------
        E : numpy.array
            Model effect evaluated at doses d.
            
        """
        pass

    @abstractmethod
    def get_parameters(self, confidence_interval=95):
        """Returns a dict of the model's parameters.

        When relevant, it will also return meaningful derived parameters. For instance, MuSyC has several parameters for E, but defines a synergy parameter beta as a function of E parameters. Thus, beta will also be included.
        
        If the model was fit to data with bootstrap_iterations > 0, this will also return the specified confidence interval.
        """
        pass

    @abstractmethod
    def summary(self, confidence_interval=95, tol=0.01):
        """Summarizes the model's synergy conclusions.

        For each synergy parameters, determines whether it indicates synergy or antagonism. When the model has been fit with bootstrap_parameters>0, the best fit, lower bound, and upper bound must all agree on synergy or antagonism.
    
        Parameters
        ----------
        confidence_interval : float, optional (default=95)
            If the model was fit() with bootstrap_parameters>0, confidence_interval will be used to get the upper and lower bounds. 

        tol : float, optional (default=0.01)
            Tolerance to determine synergy or antagonism. The parameter must exceed the threshold by at least tol (some parameters, like MuSyC's alpha which is antagonistic from 0 to 1, and synergistic from 1 to inf, will be log-scaled prior to comparison with tol)

        Returns
        ----------
        summary : str
            Tab-separated string. If the model has been bootstrapped, columns are [parameter, value, (lower,upper), synergism/antagonism]. If the model has not been bootstrapped, columns are [parameter, value, synergism/antagonism].
        """
        pass

    @abstractmethod
    def _transform_params_to_fit(self, params):
        """Some parameters may be fit on nonlinear (e.g., log) scales. This method transforms linear parameters into this scale.
        """
        pass

    @abstractmethod
    def _transform_params_from_fit(self, popt):
        """Some parameters may be fit on nonlinear (e.g., log) scales. These transformed-parameters must be transformed back to a linear scale.
        """
        pass

    @abstractmethod
    def _get_n_drugs_from_params(self, params):
        """Determine the number of drugs that are being modeled, given the size of the params array.
        """
        pass

    @abstractmethod
    def _get_initial_guess(self, d, E, p0=None):
        """Gets the initial guess used for curve_fitting
        """
        pass
    
    @abstractmethod
    def _model(self, doses, *args):
        """Model for higher dimensional parametric synergy models.

        Parameters
        ----------
        doses : numpy.ndarray
            M x N ndarray, where M is the number of samples, and N is the number of drugs.

        args
            Parameters for the model.
        """
        pass

    def plotly_isosurfaces(self, d, drug_axes=[0,1,2], other_drug_slices=None, **kwargs):
        if not self._is_parameterized():
            return None
        
        d = np.asarray(d)
        mask = d[:,0]>0
        n = d.shape[1]
        for i in range(n):
            if i in drug_axes:
                continue
            if other_drug_slices is None:
                dslice = np.min(d[:,i])
            else:
                dslice = other_drug_slices[i]
            mask = mask & (d[:,i]==dslice)

        E = self.E(d[mask,:])

        d1 = d[mask,drug_axes[0]]
        d2 = d[mask,drug_axes[1]]
        d3 = d[mask,drug_axes[2]]

        plots.plotly_isosurfaces(d1, d2, d3, E, **kwargs)



import setuptools

import codecs
import os.path

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as fh:
    long_description = fh.read()

def get_synergy_packages():
    packages=setuptools.find_packages(where='src')
    print("Hello", packages)
    return packages

setuptools.setup(
    name="synergy", # Replace with your own username
    version=get_version("src/synergy/__init__.py"),
    author="David J. Wooten",
    author_email="dwooten@psu.edu",
    description="Python package for calculating drug combination synergy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/djwooten/synergy",

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
    ],
    python_requires='>=3.5',
    keywords='synergy drug combination pharmacology cancer',
    #packages=setuptools.find_packages(where='src'),
    packages=get_synergy_packages(),
    package_dir={'': 'src'},
    install_requires=[
        "scipy >= 0.18.0", # 0.18.0 introduced curve_fit(jac=)
        "numpy >= 1.13.0" # 1.6.0 is first version compatible with python 3
        # 1.13.0 introduces np.unique(axis=) for dose_tools.get_num_replicates
    ],
    # package_data VS data_files VS ???
)



#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ..single import Hill
from .nonparametric_base import DoseDependentModel

class Schindler(DoseDependentModel):
    """Schindler's multidimensional Hill equation model.
    
    From "Theory of synergistic effects: Hill-type response surfaces as 'null-interaction' models for mixtures" - Michael Schindler. This model is built to satisfy the Loewe additivity criterion, 

    Schindler assumes each drug has a Hill dose response, and defines a multidimensional Hill equation that satisfies Loewe additivity. Unlike Loewe, Schindler can be defined for combinations whose effect exceeds Emax of either drug (Loewe is limited by Emax of the *weaker* drug).

    Synergy is simply defined as the difference between the observed E and the E predicted by Schindler's multidimensional Hill equation.

    synergy : array_like, float
        (-inf,0)=antagonism, (0,inf)=synergism
    """
    
    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, **kwargs):
        
        d1 = np.asarray(d1)
        d2 = np.asarray(d2)
        E = np.asarray(E)
        super().fit(d1,d2,E)
        
        if drug1_model is None:
            mask = np.where(d2==min(d2))
            drug1_model = Hill.create_fit(d1[mask], E[mask], E0_bounds=self.E0_bounds, Emax_bounds=self.E1_bounds, h_bounds=self.h1_bounds, C_bounds=self.C1_bounds)
        if drug2_model is None:
            mask = np.where(d1==min(d1))
            drug2_model = Hill.create_fit(d2[mask], E[mask], E0_bounds=self.E0_bounds, Emax_bounds=self.E2_bounds, h_bounds=self.h2_bounds, C_bounds=self.C2_bounds)
        
        self.drug1_model = drug1_model
        self.drug2_model = drug2_model

        E0_1, E1, h1, C1 = self.drug1_model.get_parameters()
        E0_2, E2, h2, C2 = self.drug2_model.get_parameters()
        E0 = (E0_1+E0_2)/2.
        uE1 = E0-E1
        uE2 = E0-E2
        
        uE = E0-E

        # Where d1==0 and d2==0, schindler will divide by 0
        with np.errstate(divide='ignore', invalid='ignore'):
            uE_model = self._model(d1, d2, uE1, uE2, h1, h2, C1, C2)

        self.synergy = uE-uE_model

        self.synergy[(d1==0) | (d2==0)] = 0.
        return self.synergy
    

    def _model(self, d1, d2, E1, E2, h1, h2, C1, C2):
        """
        From "Theory of synergistic effects: Hill-type response surfaces as 'null-interaction' models for mixtures" - Michael Schindler
        
        E - u_hill = 0 : Additive
        E - u_hill > 0 : Synergistic
        E - u_hill < 0 : Antagonistic
        """
        m1 = (d1/C1)
        m2 = (d2/C2)
        
        y = (h1*m1 + h2*m2) / (m1+m2)
        u_max = (E1*m1 + E2*m2) / (m1+m2)
        power = np.power(m1+m2, y)
        
        return u_max * power / (1. + power)



#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from . import base as utils

matplotlib_import = False
try:
    from matplotlib import pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    from matplotlib.cm import get_cmap
    
    matplotlib_import = True
    
except ImportError as e:
    pass

plotly_import = False
try:
    import plotly.graph_objects as go
    from plotly import offline
    plotly_import = True
except ImportError as e:
    pass

def plot_colormap(d1, d2, E, ax=None, fname=None, title="", xlabel="", ylabel="", figsize=None, cmap="PRGn_r", aspect='equal', vmin=None, vmax=None, center_on_zero=False, logscale=True, nancolor="#BBBBBB", **kwargs):
        if (not matplotlib_import):
            raise ImportError("matplotlib must be installed to plot")
        
        if logscale:
            d1 = utils.remove_zeros(d1)
            d2 = utils.remove_zeros(d2)
        else:
            d1 = np.array(d1, copy=True)
            d2 = np.array(d2, copy=True)
        E = np.asarray(E)


        sorted_indices = np.lexsort((d1,d2))
        D1 = d1[sorted_indices]
        D2 = d2[sorted_indices]
        E = E[sorted_indices]
        
        n_d1 = len(np.unique(D1))
        n_d2 = len(np.unique(D2))
        

        if len(d1) != n_d1*n_d2:
            raise ValueError("plot_colormap() requires d1, d2 to represent a dose grid")
        
        created_ax = False
        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
            created_ax=True


        if center_on_zero:
            if vmin is None or vmax is None:
                zmax = max(abs(np.nanmin(E)), abs(np.nanmax(E)))
                vmin = -zmax
                vmax = zmax
            else:
                zmax = max(abs(vmin), abs(vmax))
                vmin = -zmax
                vmax = zmax


        current_cmap = get_cmap(name=cmap)
        current_cmap.set_bad(color=nancolor)

        if not logscale:
            pco = ax.pcolormesh(D1.reshape(n_d2,n_d1), D2.reshape(n_d2, n_d1), E.reshape(n_d2, n_d1), vmin=vmin, vmax=vmax, cmap=cmap)
        else:
            pco = ax.pcolormesh(E.reshape(n_d2, n_d1), cmap=cmap, vmin=vmin, vmax=vmax)
            relabel_log_ticks(ax, np.unique(D1), np.unique(D2))

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size=max(2/n_d1, 2/n_d2, 0.05), pad=0.1)
        plt.colorbar(pco, cax=cax)
    
        ax.set_aspect(aspect)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(title)
        
        if (fname is not None):
            plt.tight_layout()
            plt.savefig(fname)
            plt.close()
        elif (created_ax):
            plt.tight_layout()
            plt.show()



def square_log_axes(ax, nx, ny):
    """
    Transform log-scaled axes into a square aspect ratio.
    The axes will be resized so that each cell in the heatmap is square.
    TODO: This will work, but only if tight_layout() is not called
    """
    ratio = ny / nx

    pos1 = ax.get_position()

    fig_width = ax.get_figure().get_figwidth()
    fig_height = ax.get_figure().get_figheight()
    axratio = (fig_height*pos1.height)/(fig_width*pos1.width)
    
    pos2 = [pos1.x0, pos1.y0,  pos1.width, pos1.height]   

    if (axratio > ratio):
        # decrease height
        target_height = ratio*fig_width*pos1.width/fig_height
        delta_height = pos1.height - target_height
        pos2[1] = pos2[1]-delta_height/2. # Move y0 up by delta/2
        pos2[3] = target_height
        ax.set_position(pos2)

    elif (axratio < ratio):
        # decrease width
        target_width = fig_height*pos1.height/(fig_width*ratio)
        #target_width = pos1.width/10
        delta_width = pos1.width - target_width
        pos2[0] = pos2[0]+delta_width/2. # Move x0 right by delta/2
        pos2[2] = target_width
        #pos2 = [0,0,1,1]
        ax.set_position(pos2)

def relabel_log_ticks(ax, d1, d2):
    """
    In plotting using pcolormesh(E), the x and y axes go from 0 to nx (or ny).
    This function replaces those with tick marks reflecting the true doses.
    Assumes both x and y axes come from log-scaled doses
    """
    nx = len(d1)
    ny = len(d2)

    MIN_logx = np.log10(min(d1))
    MAX_logx = np.log10(max(d1))
    min_logx = int(np.ceil(np.log10(min(d1))))
    max_logx = int(np.floor(np.log10(max(d1))))


    MIN_logy = np.log10(min(d2))
    MAX_logy = np.log10(max(d2))
    min_logy = int(np.ceil(np.log10(min(d2))))
    max_logy = int(np.floor(np.log10(max(d2))))

    doses = np.arange(min_logx, max_logx+1, 1)
    ticks = np.interp(doses,[MIN_logx,MAX_logx],[0.5,nx-0.5]) 
    ticklabels = [r"$10^{{{}}}$".format(dose) for dose in doses]

    ax.set_xticks(ticks)
    ax.set_xticklabels(ticklabels)

    minor_ticks = []
    for i in range(min_logx-1, max_logx+1):
        for j in range(2,10):
            minor_ticks.append(i+np.log10(j))
    minor_ticks = interp(minor_ticks,MIN_logx,MAX_logx,0.5,nx-0.5)
    minor_ticks = [i for i in minor_ticks if i>0 and i<nx]

    ax.set_xticks(minor_ticks, minor=True)




    doses = np.arange(min_logy, max_logy+1, 1)
    ticks = np.interp(doses,[MIN_logy,MAX_logy],[0.5,ny-0.5]) 
    ticklabels = [r"$10^{{{}}}$".format(dose) for dose in doses]

    ax.set_yticks(ticks)
    ax.set_yticklabels(ticklabels)


    minor_ticks = []
    for i in range(min_logy-1, max_logy+1):
        for j in range(2,10):
            minor_ticks.append(i+np.log10(j))
    minor_ticks = interp(minor_ticks,MIN_logy,MAX_logy,0.5,ny-0.5)
    minor_ticks = [i for i in minor_ticks if i>0 and i<ny]

    ax.set_yticks(minor_ticks, minor=True)

def interp(x, x0, x1, y0, y1):
    return (np.asarray(x)-x0)*(y1-y0)/(x1-x0)+y0

def plot_surface_plotly(d1, d2, E, scatter_points=None,       \
                 elev=20, azim=19, fname="plot.html", zlim=None, cmap='viridis', logscale=True,         \
                 xlabel="x", ylabel="y", zlabel="z", \
                 vmin=None, vmax=None, auto_open=True, opacity=0.8):
    if (not plotly_import):
        raise ImportError("plot_surface_plotly() requires plotly to be installed.")
    
    d1 = np.array(d1, copy=True)
    d2 = np.array(d2, copy=True)
    E = np.asarray(E)


    if logscale:
        d1 = utils.remove_zeros(d1)
        d2 = utils.remove_zeros(d2)
        d1 = np.log10(d1)
        d2 = np.log10(d2)

    sorted_indices = np.lexsort((d1,d2))
    d1 = d1[sorted_indices]
    d2 = d2[sorted_indices]
    E = E[sorted_indices]

    n_d1 = len(np.unique(d1))
    n_d2 = len(np.unique(d2))
    d1 = d1.reshape(n_d2,n_d1)
    d2 = d2.reshape(n_d2,n_d1)
    E = E.reshape(n_d2,n_d1)

    data_to_plot = [go.Surface(
        x=d1,
        y=d2,
        z=E,
        cmin=vmin,
        cmax=vmax,
        opacity=opacity,
        contours_z=dict(
            show=True,
            usecolormap=True,
            highlightcolor="limegreen",
            project_z=False
        ),
        colorscale=cmap,
        reversescale=True,
        colorbar=dict(
            lenmode='fraction',
            len=0.65,
            title=zlabel
            )
        ),
    ]

    if scatter_points is not None:
        #d1scatter = utils.remove_zeros(np.asarray(scatter_points['drug1.conc']))
        #d2scatter = utils.remove_zeros(np.asarray(scatter_points['drug2.conc']))
        d1scatter = np.array(scatter_points['drug1.conc'], copy=True)
        d2scatter = np.array(scatter_points['drug2.conc'], copy=True)
        if logscale:
            zero_mask_1 = np.where(d1scatter <= 0)
            pos_mask_1 = np.where(d1scatter > 0)

            zero_mask_2 = np.where(d2scatter <= 0)
            pos_mask_2 = np.where(d2scatter > 0)

            d1scatter[zero_mask_1] = np.min(d1)
            d2scatter[zero_mask_2] = np.min(d2)
            d1scatter[pos_mask_1] = np.log10(d1scatter[pos_mask_1])
            d2scatter[pos_mask_2] = np.log10(d2scatter[pos_mask_2])
        
        data_to_plot.append(go.Scatter3d(
            x=d1scatter,
            y=d2scatter,
            z=scatter_points['effect'],
            mode="markers",
            marker=dict(
                size=1.5,
                color=scatter_points['effect'],
                colorscale=cmap,
                reversescale=True,
                cmin=vmin,
                cmax=vmax,
                line=dict(
                    width=0.5,
                    color='DarkSlateGrey'
                )
            )
        ))

    fig = go.Figure(data=data_to_plot)

    #fig.update_traces(contours_z=dict(show=True, usecolormap=True, highlightcolor="limegreen", project_z=True))

    title = ""
    if fname is not None:
        title = fname

    fig.update_layout(
        title=title,
        autosize=False,
        scene_camera_eye=dict(
            x=1.87, 
            y=0.88, 
            z=0.64
        ), 
        width=1000, 
        height=800, 
        margin=dict(
            l=100, 
            r=100, 
            b=90, 
            t=90
        ), 
        scene=dict(
            xaxis_title=xlabel, 
            yaxis_title=ylabel, 
            zaxis_title=zlabel, 
            aspectmode="cube"
        )
    )

    if zlim is not None:
        fig.update_layout(scene = dict(zaxis = dict(range=zlim,)))


    if fname is not None:
        offline.plot(fig, filename=fname, auto_open=auto_open)
    else:
        fig.show()


def plotly_isosurfaces(d1, d2, d3, E, fname=None,     \
            cmap='viridis', xlabel="x", ylabel="y", zlabel="z",     \
            vmin=None, vmax=None, auto_open=True, opacity=0.6,      \
            logscale=True, isomin=None, isomax=None,                \
            center_on_zero=False, surface_count=10):
    
    d1 = np.asarray(d1)
    d2 = np.asarray(d2)
    d3 = np.asarray(d3)
    
    if (len(d1.shape) > 1):
        d1 = d1.flatten()
        d2 = d2.flatten()
        d3 = d3.flatten()
        E = E.flatten()

    sorted_indices = np.lexsort((d1,d2,d3))
    d1 = d1[sorted_indices]
    d2 = d2[sorted_indices]
    d3 = d3[sorted_indices]
    E = E[sorted_indices]

    if logscale:
        d1 = utils.remove_zeros(d1)
        d2 = utils.remove_zeros(d2)
        d3 = utils.remove_zeros(d3)
        d1 = np.log10(d1)
        d2 = np.log10(d2)
        d3 = np.log10(d3)
    
    
    E_range = np.nanmax(E[~np.isinf(E)]) - np.nanmin(E[~np.isinf(E)])
    if isomin is None:
        isomin = np.nanmin(E[~np.isinf(E)]) + 0.1*E_range
    if isomax is None:
        isomax = np.nanmin(E[~np.isinf(E)]) + 0.9*E_range

    if center_on_zero:
        if vmin is None or vmax is None:
            zmax = max(abs(np.nanmin(E[~np.isinf(E)])), abs(np.nanmax(E[~np.isinf(E)])))
            vmin = -zmax
            vmax = zmax
        else:
            zmax = max(abs(vmin), abs(vmax))
            vmin = -zmax
            vmax = zmax

        
    fig = go.Figure(data=go.Isosurface(
        x=d1,
        y=d2,
        z=d3,
        value=E,
        isomin=isomin,
        isomax=isomax,
        cmin=vmin,
        cmax=vmax,
        opacity=0.6,
        colorscale='PRGn_r',
        surface_count=surface_count, # number of isosurfaces, 2 by default: only min and max
        colorbar_nticks=surface_count, # colorbar ticks correspond to isosurface values
        caps=dict(x_show=False, y_show=False, z_show=True)
    ))

    title = ""
    if fname is not None:
        title = fname

    fig.update_layout(
        title=title,
        autosize=False,
        scene_camera_eye=dict(
            x=1.87, 
            y=0.88, 
            z=0.64
        ), 
        width=1000, 
        height=800, 
        margin=dict(
            l=100, 
            r=100, 
            b=90, 
            t=90
        ), 
        scene=dict(
            xaxis_title=xlabel, 
            yaxis_title=ylabel, 
            zaxis_title=zlabel, 
            aspectmode="cube"
        )
    )

    if fname is not None:
        offline.plot(fig, filename=fname, auto_open=auto_open)
    else:
        fig.show()


class ModelNotParameterizedError(Exception):
    """
    The model must be parameterized prior to use. This can be done by calling
    fit(), or setParameters().
    """
    def __init__(self, msg='The model must be parameterized prior to use. This can be done by calling fit(), or setParameters().', *args, **kwargs):
        super().__init__(msg, *args, **kwargs)

class FeatureNotImplemented(Warning):
    def __init__(self, msg="This feature is not yet implemented", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)



#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.stats import linregress

from .parametric_base import ParameterizedModel1D
from .. import utils

class Hill(ParameterizedModel1D):
    """The four-parameter Hill equation

                            d^h
    E = E0 + (Emax-E0) * ---------
                         C^h + d^h

    The Hill equation is a standard model for single-drug dose-response curves.
    This is the base model for Hill_2P and Hill_CI.

    """
    def __init__(self, E0=None, Emax=None, h=None, C=None, E0_bounds=(-np.inf, np.inf), Emax_bounds=(-np.inf, np.inf), h_bounds=(0,np.inf), C_bounds=(0,np.inf)):
        """
        Parameters
        ----------
        E0 : float, optional
            Effect at 0 dose. Set this if you are creating a synthetic Hill
            model, rather than fitting from data
        
        Emax : float, optional
            Effect at 0 dose. Set this if you are creating a synthetic Hill
            model, rather than fitting from data
        
        h : float, optional
            The Hill-slope. Set this if you are creating a synthetic Hill
            model, rather than fitting from data
        
        C : float, optional
            EC50, the dose for which E = (E0+Emax)/2. Set this if you are
            creating a synthetic Hill model, rather than fitting from data
        
        X_bounds: tuple
            Bounds to use for Hill equation parameters during fitting. Valid options are E0_bounds, Emax_bounds, h_bounds, C_bounds.
        """

        super().__init__()
        self.E0 = E0
        self.Emax = Emax
        self.h = h
        self.C = C
        
        self.E0_bounds=E0_bounds
        self.Emax_bounds=Emax_bounds
        self.h_bounds=h_bounds
        self.C_bounds=C_bounds
        with np.errstate(divide='ignore'):
            self.logh_bounds = (np.log(h_bounds[0]), np.log(h_bounds[1]))
            self.logC_bounds = (np.log(C_bounds[0]), np.log(C_bounds[1]))

        self.fit_function = lambda d, E0, E1, logh, logC: self._model(d, E0, E1, np.exp(logh), np.exp(logC))

        self.jacobian_function = lambda d, E0, E1, logh, logC: self._model_jacobian(d, E0, E1, logh, logC)

        self.bounds = tuple(zip(self.E0_bounds, self.Emax_bounds, self.logh_bounds, self.logC_bounds))

    def E(self, d):
        """Evaluate this model at dose d. If the model is not parameterized, returns 0.

        Parameters
        ----------
        d : array_like
            Doses to calculate effect at
        
        Returns
        ----------
        effect : array_like
            Evaluate's the model at dose in d
        """
        if not self._is_parameterized():
            return super().E(d)

        return self._model(d, self.E0, self.Emax, self.h, self.C)

    def E_inv(self, E):
        """Inverse of the Hill equation

        Parameters
        ----------
        E : array_like
            Effects to get the doses for
        
        Returns
        ----------
        doses : array_like
            Doses which achieve effects E using this model. Effects that are
            outside the range [E0, Emax] will return np.nan for the dose
        """
        if not self._is_parameterized():
            return super().E_inv(E)

        return self._model_inv(E, self.E0, self.Emax, self.h, self.C)

    def get_parameters(self):
        """Gets the model's parmaters.

        Returns
        ----------
        parameters : tuple
            (E0, Emax, h, C)
        """
        return (self.E0, self.Emax, self.h, self.C)

    def _set_parameters(self, popt):
        E0, Emax, h, C = popt
        
        self.E0 = E0
        self.Emax = Emax
        self.h = h
        self.C = C
        
    def _model(self, d, E0, Emax, h, C):
        dh = d**h
        return E0 + (Emax-E0)*dh/(C**h+dh)

    def _model_inv(self, E, E0, Emax, h, C):
        d = np.float_power((E0-Emax)/(E-Emax)-1.,1./h)*C
        return d

    def _model_jacobian(self, d, E0, Emax, logh, logC):
        """
        Returns
        ----------
        jacobian : array_like
            Derivatives of the Hill equation with respect to E0, Emax, logh,
            and logC
        """
        
        dh = d**(np.exp(logh))
        Ch = (np.exp(logC))**(np.exp(logh))
        logd = np.log(d)

        jE0 = 1 - dh/(Ch+dh)
        jEmax = 1-jE0

        jC = (E0-Emax)*dh*np.exp(logh+logC)*(np.exp(logC))**(np.exp(logh)-1) / ((Ch+dh)*(Ch+dh))

        jh = (Emax-E0)*dh*np.exp(logh) * ((Ch+dh)*logd - (logC*Ch + logd*dh)) / ((Ch+dh)*(Ch+dh))
        
        jac = np.hstack((jE0.reshape(-1,1), jEmax.reshape(-1,1), jh.reshape(-1,1), jC.reshape(-1,1)))
        jac[np.isnan(jac)]=0
        return jac

    def _get_initial_guess(self, d, E, p0=None):

        if p0 is None:
            p0 = [max(E), min(E), 1, np.median(d)]

        p0 = list(self._transform_params_to_fit(p0))
        utils.sanitize_initial_guess(p0, self.bounds)
        return p0

    def _transform_params_from_fit(self, params):
        return params[0], params[1], np.exp(params[2]), np.exp(params[3])

    def _transform_params_to_fit(self, params):
        return params[0], params[1], np.log(params[2]), np.log(params[3])

    def create_fit(d, E, E0_bounds=(-np.inf, np.inf), Emax_bounds=(-np.inf, np.inf), h_bounds=(0,np.inf), C_bounds=(0,np.inf), **kwargs):
        """Courtesy function to build a Hill model directly from data.
        Initializes a model using the provided bounds, then fits.
        """
        drug = Hill(E0_bounds=E0_bounds, Emax_bounds=Emax_bounds, h_bounds=h_bounds, C_bounds=C_bounds)
        drug.fit(d, E, **kwargs)
        return drug

    def __repr__(self):
        if not self._is_parameterized(): return "Hill()"
        
        return "Hill(E0=%0.2f, Emax=%0.2f, h=%0.2f, C=%0.2e)"%(self.E0, self.Emax, self.h, self.C)












class Hill_2P(Hill):
    """The two-parameter Hill equation

                            d^h
    E = E0 + (Emax-E0) * ---------
                         C^h + d^h

    Mathematically equivalent to the four-parameter Hill equation, but E0 and Emax are held constant (not fit to data).
    
    """
    def __init__(self, h=None, C=None, h_bounds=(0,np.inf), C_bounds=(0,np.inf), E0=1, Emax=0):
        super().__init__(h=h, C=C, E0=E0, Emax=Emax, h_bounds=h_bounds, C_bounds=C_bounds)

        self.fit_function = lambda d, logh, logC: self._model(d, self.E0, self.Emax, np.exp(logh), np.exp(logC))

        self.jacobian_function = lambda d, logh, logC: self._model_jacobian(d, logh, logC)

        self.bounds = tuple(zip(self.logh_bounds, self.logC_bounds))

    def _model_jacobian(self, d, logh, logC):
        dh = d**(np.exp(logh))
        Ch = (np.exp(logC))**(np.exp(logh))
        logd = np.log(d)
        E0 = self.E0
        Emax = self.Emax

        jC = (E0-Emax)*dh*np.exp(logh+logC)*(np.exp(logC))**(np.exp(logh)-1) / ((Ch+dh)*(Ch+dh))

        jh = (Emax-E0)*dh*np.exp(logh) * ((Ch+dh)*logd - (logC*Ch + logd*dh)) / ((Ch+dh)*(Ch+dh))
        
        jac = np.hstack((jh.reshape(-1,1), jC.reshape(-1,1)))
        jac[np.isnan(jac)]=0
        return jac

    def _get_initial_guess(self, d, E, p0=None):

        if p0 is None:
            p0 = [1, np.median(d)]
            
        p0 = list(self._transform_params_to_fit(p0))
        utils.sanitize_initial_guess(p0, self.bounds)
        
        return p0

    def create_fit(d, E, E0=1, Emax=0, h_bounds=(0,np.inf), C_bounds=(0,np.inf), **kwargs):
        drug = Hill_2P(E0=E0, Emax=Emax, h_bounds=h_bounds, C_bounds=C_bounds)
        drug.fit(d, E, **kwargs)
        return drug

    def get_parameters(self):
        """Gets the model's parameters
        
        Returns
        ----------
        parameters : tuple
            (h, C)
        """
        return (self.h, self.C)

    def _set_parameters(self, popt):
        h, C = popt
        
        self.h = h
        self.C = C

    def _transform_params_from_fit(self, params):
        return np.exp(params[0]), np.exp(params[1])

    def _transform_params_to_fit(self, params):
        return np.log(params[0]), np.log(params[1])

    def __repr__(self):
        if not self._is_parameterized(): return "Hill()"
        
        return "Hill2P(E0=%0.2f, Emax=%0.2f, h=%0.2f, C=%0.2e)"%(self.E0, self.Emax, self.h, self.C)
    
    







class Hill_CI(Hill_2P):
    """Mathematically equivalent two-parameter Hill equation with E0=1 and Emax=0. However, Hill_CI.fit() uses the log-linearization approach to dose-response fitting used by the Combination Index.
    """
    def __init__(self, h=None, C=None):
        super().__init__(h=h, C=C, E0=1., Emax=0.)


    def _internal_fit(self, d, E, use_jacobian, **kwargs):
        mask = np.where((E < 1) & (E > 0) & (d > 0))
        E = E[mask]
        d = d[mask]
        fU = E
        fA = 1-E

        median_effect_line = linregress(np.log(d),np.log(fA/fU))
        h = median_effect_line.slope
        C = np.exp(-median_effect_line.intercept / h)
        
        return (h, C)

    def create_fit(d, E):
        drug = Hill_CI()
        drug.fit(d, E)
        return drug

    def plot_linear_fit(self, d, E, ax=None):
        if not self._is_parameterized():
            # TODO: Error
            return

        try:
            from matplotlib import pyplot as plt
        except:
            # TODO: Error
            # TODO: Move this whole function to utils.plot
            return
        mask = np.where((E < 1) & (E > 0) & (d > 0))
        E = E[mask]
        d = d[mask]
        fU = E
        fA = 1-E

        ax_created = False
        if ax is None:
            ax_created = True
            fig = plt.figure(figsize=(4,4))
            ax = fig.add_subplot(111)

        ax.scatter(np.log(d), np.log(fA/fU))
        ax.plot(np.log(d), np.log(d)*self.h - self.h*np.log(self.C))

        if False:
            for i in range(self.bootstrap_parameters.shape[0]):
                h, C = self.bootstrap_parameters[i,:]
                ax.plot(np.log(d), np.log(d)*h - h*np.log(C), c='k', alpha=0.1, lw=0.5)

        ax.set_ylabel("h*log(d) - h*log(C)")
        ax.set_xlabel("log(d)")
        ax.set_title("CI linearization")
        if (ax_created):
            plt.tight_layout()
            plt.show()

    def _bootstrap_resample(self, d, E, use_jacobian, bootstrap_iterations, confidence_interval, **kwargs):
        """Bootstrap resampling is not yet implemented for CI
        """
        pass

    def __repr__(self):
        if not self._is_parameterized(): return "Hill_CI()"
        
        return "Hill_CI(h=%0.2f, C=%0.2e)"%(self.h, self.C)
    
    